from wtforms import Form, StringField, PasswordField, RadioField, IntegerField, HiddenField, FieldList, FormField, BooleanField, DateTimeField, TextAreaField, SelectField
from wtforms import validators
from wtforms.fields.html5 import EmailField
from wtforms.validators import DataRequired, Optional


class RequiredIf(DataRequired):
    """Validator which makes a field required if another field is set and has a truthy value.

    Sources:
        - http://wtforms.simplecodes.com/docs/1.0.1/validators.html
        - http://stackoverflow.com/questions/8463209/how-to-make-a-field-conditionally-optional-in-wtforms
        - https://gist.github.com/devxoul/7638142#file-wtf_required_if-py
    """
    field_flags = ('requiredif',)

    def __init__(self, message=None, *args, **kwargs):
        super(RequiredIf).__init__()
        self.message = message
        self.conditions = kwargs

    # field is requiring that name field in the form is data value in the form
    def __call__(self, form, field):
        for name, data in self.conditions.items():
            other_field = form[name]
            if other_field is None:
                raise Exception('no field named "%s" in form' % name)
            if other_field.data == data and not field.data:
                DataRequired.__call__(self, form, field)
            Optional()(form, field)

class PassCodeForm(Form):
    passcode = PasswordField('Passcode',[validators.DataRequired()],description='Enter passcode to change mongo database details')

class AdminForm(Form):
    hostname = StringField('Database Hostname', [validators.DataRequired("Please enter Hostname of the mongo database.")],description='Hostname of the mongo database',render_kw={"placeholder": "Enter hostname to connect to database"})
    port = IntegerField('Port No.',description='Port No of database',render_kw={"placeholder": "Enter port number to connect to database"})
    username = StringField('Username', [validators.DataRequired("Please enter username to connect to database.")],description='Authorized user name to connect database',render_kw={"placeholder": "Enter username to connect to database"})
    password = PasswordField('Password', [validators.DataRequired("Please enter password to connect to database.")],description='Password to connect to database',render_kw={"placeholder": "Enter password to connect to database "})
    dbname = StringField('Database Name', [validators.DataRequired("Please enter name of database.")],description='Name of mongo database',render_kw={"placeholder": "Enter name of mongo database"})
    collection = StringField('Collection', [validators.DataRequired("Please enter collection of database.")],description='Collection of mongo database',render_kw={"placeholder": "Enter collection of database"})
    isSSL = RadioField('IsSSL?', choices=[('Yes', 'Yes'), ('No', 'No')], default='No',description='Is database SSL protected?')

class ConfigForm(Form):
    httpService = RadioField('Is there a HTTP service running on the server?', choices=[('Yes', 'Yes'), ('No', 'No')], default='No', description='e.g. https://notebook.rcc.uchicago.edu/files')
    gitService = RadioField('Is there a Git service running on the server?',choices=[('Yes', 'Yes'), ('No', 'No')], default='No')
    globusService = RadioField('Is there a Globus service running on the server?',choices=[('Yes', 'Yes'), ('No', 'No')], default='No',description='e.g. https://www.globus.org/app/transfer?origin_id=72277ed4-1ad3-11e7-bbe1-22000b9a448b&origin_path=')
    downloadPath = StringField('Download Path',description='The Globus service allows to download the paper content using gridFTP.')
    fileServerPath = StringField('File Server Path', description='If a HTTP service is running on the server, a URL may be associated to the paper content. This URL is only required by Qresp | Exploration to view images & download files using a web browser.')

class NameForm(Form):
    firstName = StringField(validators=[validators.DataRequired()],description='e.g. John',render_kw={"placeholder": "Enter first name"})
    middleName = StringField(description='e.g. L.',render_kw={"placeholder": "Enter middle name"})
    lastName = StringField(validators=[validators.DataRequired()], description='e.g. Doe',render_kw={"placeholder": "Enter last name"})

class DetailsForm(Form):
    firstName = StringField('First Name',validators=[validators.DataRequired()],description='e.g. John',render_kw={"placeholder": "Enter first name"})
    middleName = StringField('Middle Name',description='e.g. L.',render_kw={"placeholder": "Enter middle name"})
    lastName = StringField('Last Name',validators=[validators.DataRequired()], description='e.g. Doe',render_kw={"placeholder": "Enter last name"})
    emailId = EmailField('Email Address', description='e.g. john.doe@company.com',render_kw={"placeholder": "Enter your email address"})
    affiliation = StringField('Affiliation', description='e.g. Department of Chem, University of XYZ',render_kw={"placeholder": "Enter your university"})

class ServerForm(Form):
    serverName = StringField('Server Name', [validators.DataRequired("Please enter hostname of server")], description='Enter Hostname of remote server where paper content is organized and located e.g. midway.rcc.uchicago.edu',render_kw={"placeholder": "Enter hostname of server"})
    username = StringField('Username', [validators.DataRequired("Please enter username of server")], description='Enter your username to establish a SSH connection to the remote server',render_kw={"placeholder": "Enter username to login"})
    password = PasswordField('Password', [validators.DataRequired("Please enter password of server")], description='Enter your password associated to the username',render_kw={"placeholder": "Enter password to login"})
    isDUOAuth = RadioField('Is 2FA DUO authorized?', choices=[('Yes', 'Yes'), ('No', 'No')], default='Yes',description='Is server 2FA DUO authorized?')


class ProjectForm(Form):
    downloadPath = HiddenField()
    fileServerPath = HiddenField()
    notebookPath = HiddenField()
    folderAbsolutePath = StringField('Folder Absolute Path', [validators.DataRequired()], description='e.g. /home/public',render_kw={"placeholder": "Enter relative path to the data"})
    ProjectName = HiddenField()
    gitPath = HiddenField()
    insertedBy = FormField(DetailsForm)
    isPublic = BooleanField(default=True)
    timeStamp = DateTimeField()
    serverPath = HiddenField()
    notebookFile = HiddenField()
    doi = HiddenField()

class InfoForm(Form):
    PIs = FieldList(FormField(NameForm), label='Principal Investigator(s)', description='Enter PI(s)', min_entries=1, validators=[validators.DataRequired()])
    collections = StringField(label='PaperStack',validators = [validators.DataRequired("Please enter keywords")], description='Enter name(s) defining group of papers (e.g. according to source of fundings)',render_kw={"placeholder": "Enter collection to which project belongs"})
    tags = StringField(label='Keywords',validators = [validators.DataRequired("Please enter keywords")], description='Enter Keyword(s) (e.g: DFT, organic materials, charge transfer): they facilitate paper searches using Qresp | Explorer',render_kw={"placeholder": "Enter tags for the project"})
    mainnotebookfile = StringField('Main Notebook File', description='Enter name of a notebook file, this file may serve as a table of contents and may contain links to all datasets, charts, scripts, tools and documentation. Use the Paper Content widget (on the left) to fill this field',render_kw={"placeholder": "Enter main notebook filename"})
    doi = StringField('DOI',description='DOI minted by Qresp',render_kw={'readonly': True})

class ExtraForm(Form):
    extrakey = StringField(description='Enter custom label',render_kw={"placeholder": "Enter Field Label"})
    extravalue = StringField(description='Enter custom value', render_kw={"placeholder": "Enter Field Value"})

class ChartForm(Form):
    id = HiddenField('Id')
    caption = StringField('Caption',[validators.DataRequired("Please enter caption")], description='Enter chart caption',render_kw={"placeholder": "Enter chart caption"})
    number = StringField('Number',[validators.DataRequired("Please enter number")], description='Enter chart number',render_kw={"placeholder": "Enter chart number"})
    files = StringField(label='Files', description='Enter file name(s) containing the data displayed in the chart (e.g. a file in CSV format). Use the Paper Content widget (on the left) to fill this field',render_kw={"placeholder": "Enter file names used to construct Chart"})
    imageFile = StringField('Image File', [validators.DataRequired("Please enter image file")], description='Enter the file name containing a snapshot of the chart. Use the Paper Content widget (on the left) to fill this field. Allowed formats are: jpg, jpeg, gif, png',render_kw={"placeholder": "Enter Image file name"})
    notebookFile = StringField('Notebook File', description='Enter the name of the notebook file used to generate the chart. Use the Paper Content widget (on the left) to fill this field. Allowed format is ipynb',render_kw={"placeholder": "Enter Notebook file"})
    properties = StringField(label='Keywords',validators= [validators.DataRequired("Please enter properties")], description='Enter keyword(s) for the content displayed in the chart. e.g. potential energy surface, band gap',render_kw={"placeholder": "Enter Keywords"})
    saveas = StringField('Save As', [validators.DataRequired()], description='Enter a name to identify the chart', render_kw={"placeholder": "Save Chart as"})
    extraFields = FieldList(FormField(ExtraForm), label='Extra Fields', description='Enter a label name and add a value to it',min_entries=1,id='extraChartFields')


class ToolForm(Form):
    id = HiddenField('Id')
    kind = RadioField('Kind', choices=[('software', 'Software'), ('experiment', 'Experiment')], default='software',description='Select Software or Experiment')
    packageName = StringField('Package Name', validators = [RequiredIf(kind='Software')], description='Enter name of the package (e.g. WEST)',render_kw={"placeholder": "Enter Package name for software tools"})
    URLs = StringField(label='URLs', description='Enter link(s) to package official website (e.g. www.west-code.org)',render_kw={"placeholder": "Enter Urls for software tools"})
    version = StringField('Version',  validators = [RequiredIf(kind='Software')], description='Enter version number (e.g. 3.0.0) of the package',render_kw={"placeholder": "Enter version for software tools"})
    programName = StringField('Executable Name', description='Enter name of the executable (e.g. wstat.x) of the package',render_kw={"placeholder": "Enter Program name for software tools"})
    patches = StringField(label='Patches', description='Enter file name(s) containing the patches of publicly available or versioned software, customized by the user to generate some of the datasets. Use the Paper Content widget (on the left) to fill this field. (e.g. Tools/modified_wstat.txt)',render_kw={"placeholder": "Enter files for software tool"})
    description = StringField('Description', description='Enter a summary of the modifications, if any, made to the executable', render_kw={"placeholder": "Enter Description for script added to the Software Tool"})
    facilityName = StringField('Facility Name',  validators = [RequiredIf(kind='Experiment')], description='Enter name of the facility where the experiment was conducted (e.g. Argonne Advanced Photon Source)', render_kw={"placeholder": "Enter Facility name for experiment tools"})
    measurement = StringField('Measurement',  validators = [RequiredIf(kind='Experiment')], description='Enter type of measurement (e.g. soft X-ray Photoemission)', render_kw={"placeholder": "Enter measurement for experiment tools"})
    saveas = StringField('Save As', [validators.DataRequired()], description='Enter a name to identify the tool', render_kw={"placeholder": "Save tool as"})
    extraFields = FieldList(FormField(ExtraForm), label='Extra Fields', description='Enter a label name and add a value to it',min_entries=1,id='extraToolFields')

class DatasetForm(Form):
    id = HiddenField('Id')
    files = StringField(label='Files', validators = [validators.DataRequired()], description='Enter file name(s) to identify the dataset. Use the Paper Content widget (on the left) to fill this field (e.g. Data/dataset.dat). If you list a folder name, all documents of the folder belong to the dataset',render_kw={"placeholder": "Enter files for dataset"})
    readme = StringField('Description', [validators.DataRequired()], description='Enter a summary about the content of the dataset',render_kw={"placeholder": "Enter descriptions For dataset"})
    URLs = StringField(label='URLs', description='Enter link(s) to the URL of the dataset, if available',render_kw={"placeholder": "Enter Urls for dataset"})
    saveas = StringField('Save As', [validators.DataRequired()], description='Enter a name to identify the dataset', render_kw={"placeholder": "Save dataset as"})
    extraFields = FieldList(FormField(ExtraForm), label='Extra Fields', description='Enter a label name and add a value to it',min_entries=1,id='extraDatasetFields')

class ScriptForm(Form):
    id = HiddenField('Id')
    files = StringField(label='Files', validators = [validators.DataRequired()], description='Enter file names to identify the script. Use the Paper Content widget (on the left) to fill this field (e.g. Script/scriptA.py). If you list a folder name, all documents of the folder belong to the script',render_kw={"placeholder": "Enter files for script"})
    readme = StringField('Description', [validators.DataRequired()], description='Enter a summary about the content of the script',render_kw={"placeholder": "Enter description for script"})
    URLs = StringField(label='URLs', description='Enter link(s) to the URL of the script, if available',render_kw={"placeholder": "Enter Urls for script"})
    saveas = StringField('Save As', [validators.DataRequired()], description='Enter a name to identify the script', render_kw={"placeholder": "Save script as"})
    extraFields = FieldList(FormField(ExtraForm), label='Extra Fields', description='Enter a label name and add a value to it',min_entries=1,id='extraScriptFields')

class JournalForm(Form):
    abbrevName = StringField('Abbreviation',description='Enter short Journal Name')
    fullName = StringField('Journal Name',description='Enter full Journal Name',render_kw={"placeholder": "Enter full journal name"})


class ReferenceForm(Form):
    kind = RadioField('Kind', choices=[('preprint', 'Preprint'), ('journal', 'Journal'), ('dissertation','Dissertation')], default='journal',description='Select Preprint, Journal or Dissertation')
    DOI = StringField('DOI', [validators.DataRequired()], description='Enter DOI of paper', render_kw={"placeholder": "Enter DOI of paper"})
    authors = FieldList(FormField(NameForm),label='Authors', validators = [validators.DataRequired()], description='Enter authors of paper', render_kw={"placeholder": "Enter authors"},min_entries=1)
    title = StringField('Title', [validators.DataRequired()], description='Enter title of paper', render_kw={"placeholder": "Enter title"})
    journal = FormField(JournalForm)
    page = StringField('Page', [validators.DataRequired()], description='Enter page number of journal', render_kw={"placeholder": "Enter page number"})
    publishedAbstract = StringField('Abstract', [validators.DataRequired()], description='Enter abstract', render_kw={"placeholder": "Enter abstract"})
    volume = StringField('Volume', [validators.DataRequired()], description='Enter volume of journal', render_kw={"placeholder": "Enter volume number"})
    year = IntegerField('Year', [validators.DataRequired()], description='Enter year', render_kw={"placeholder": "Enter year"})
    URLs = StringField(label='URLs', description='Enter link(s) of the paper',render_kw={"placeholder": "Enter paper Urls"})
    school = StringField('School', description='Enter name of school where dissertation was presented', render_kw={"placeholder": "Enter name of school"})

class DocumentationForm(Form):
    readme = TextAreaField('Readme',description='Enter documentation about paper', render_kw={"placeholder": "Enter documentation of paper"})

class WorkflowForm(Form):
    edges = FieldList(FieldList(StringField()))
    nodes = FieldList(StringField())

class VersionsForm(Form):
    versions = FieldList(StringField())

class HeadForm(Form):
    id = HiddenField('Id')
    readme = StringField('Description',description='Describe your external resource', render_kw={"placeholder": "Describe external resource"})
    URLs = StringField('URLs',description='Enter URLs for the external resource')

class PublishForm(Form):
    server = SelectField('Qresp Server',choices=[], validators=[validators.DataRequired()],description='Select the Address of the Database. Qresp automatically inserts the metadata file in the database')
    emailId = EmailField('Email Address', validators = [validators.DataRequired()],description='e.g. john.doe@gmail.com',render_kw={"placeholder": "Enter a Google based email address to verify your identity."})

class PaperForm(Form):
    PIs = FieldList(FormField(NameForm))
    charts = FieldList(FormField(ChartForm))
    collections = FieldList(StringField())
    datasets = FieldList(FormField(DatasetForm))
    info = FormField(ProjectForm)
    reference = FormField(ReferenceForm)
    scripts = FieldList(FormField(ScriptForm))
    tools = FieldList(FormField(ToolForm))
    tags = FieldList(StringField())
    versions = FieldList(FormField(VersionsForm))
    workflow = FormField(WorkflowForm)
    heads = FieldList(FormField(HeadForm))
    schema = StringField()
    version = IntegerField()

class WorkflowCreator(object):
    """Class to create stand alone final workflow for the metadata(JSON) format.
    """

    def __init__(self):
        self.desc = {}
        self.desc["charts"] = []
        self.desc["datasets"] = []
        self.desc["scripts"] = []
        self.desc["tools"] = []
        self.desc["heads"] = []
        self.desc["edges"] = []
        self.desc["nodes"] = []

    def addChart(self, chartsaveas, id):
        self.desc["charts"].append(str(chartsaveas) + "*" + str(id))

    def addDataset(self, datasetsaveas, id):
        self.desc["datasets"].append(str(datasetsaveas) + "*" + str(id))

    def addScript(self, scriptsaveas, id):
        self.desc["scripts"].append(str(scriptsaveas) + "*" + str(id))

    def addTool(self, toolssaveas, id):
        self.desc["tools"].append(str(toolssaveas) + "*" + str(id))

    def addHead(self, head, id, url=None):
        if url is not None:
            self.desc["heads"].append(str(head) + "*" + str(id) + "*" + url)
        else:
            self.desc["heads"].append(str(head) + "*" + str(id))

    def addEdge(self, edge):
        self.desc["edges"].append(edge)

    def addNode(self, node):
        self.desc["nodes"].append(node)






import re
import traceback
from .db import *
from .models import *


class PaperDAO(MongoDBConnection):
    def __init__(self):
        self.hasvisited = []
        self.workflowinfo = WorkflowInfo()
        self.workflowinfo.nodes = {}
        self.workflowinfo.edges = []

    def getCollectionList(self):
        """fetches all collections from paper
        :return paperCollectionlist: list
        """
        paperCollectionlist = Paper.objects.get_unique_values('collections')
        return paperCollectionlist

    def getPublicationList(self):
        """fetches all publications from paper
            :return paperPublicationlist: list
        """
        paperPublicationlist = Paper.objects.get_unique_values('reference.journal.fullName')
        return paperPublicationlist

    def getAuthorList(self):
        """fetches all authors from paper
            :return authorslist: list
        """
        authorslist = Paper.objects.get_unique_names('reference.authors')
        return authorslist

    def getAllPapers(self):
        """
        fetches all papers
        :return:
        """
        allPapers = [paper.to_json() for paper in Paper.objects()]
        return allPapers

    def getAllFilteredSearchObjects(self, searchWord, paperTitle, doi, tags, collectionList, authorsList,
                                    publicationList):
        """
        Produces papers after filtering with word
        :param searchWord:
        :return: Filtered Paper
        """
        allFilteredSearchObjects = []
        if searchWord:
            regSearch = '.*' + searchWord + '.*'
            searchRegWord = re.compile(regSearch, re.IGNORECASE)
            filteredPaper = Paper.objects.filter(
                Q(reference__title=searchRegWord) | Q(reference__publishedAbstract=searchRegWord) |
                Q(tags=searchRegWord) | Q(collections=searchRegWord) | Q(reference__authors__firstName=searchWord) |
                Q(reference__authors__lastName=searchWord))
            allFilteredSearchObjects = self.__filtersearchedPaper(filteredPaper)
        else:
            searchquery = Q()
            if paperTitle.strip():
                searchTitleQuery = Q(reference__title__icontains=paperTitle)
                searchquery = searchquery & searchTitleQuery
            if doi.strip():
                searchdoiquery = Q(reference__DOI__icontains=doi)
                searchquery = searchquery & searchdoiquery
            if tags.strip():
                searchtagquery = Q(tags__icontains=tags)
                searchquery = searchquery & searchtagquery
            if len(collectionList) > 0:
                searchcollectionquery = Q(collections__in=collectionList)
                searchquery = searchquery & searchcollectionquery
            if len(authorsList) > 0:
                lastnamelist = [name.split(" ")[1] for name in authorsList]
                searchauthorquery = Q(reference__authors__lastName__in=lastnamelist)
                searchquery = searchquery & searchauthorquery
            if len(publicationList) > 0:
                searchpubquery = Q(reference__journal__fullName__in=publicationList)
                searchquery = searchquery & searchpubquery
            if searchquery:
                filteredPaper = Paper.objects.filter(searchquery)
                allFilteredSearchObjects = self.__filtersearchedPaper(filteredPaper)
            else:
                allFilteredSearchObjects = self.__filtersearchedPaper(Paper.objects())
        return allFilteredSearchObjects

    def getAllSearchObjects(self):
        """ Produces all search objects
        :return: list: all Search objects
        """
        allSearchobjects = self.__filtersearchedPaper(Paper.objects())
        return allSearchobjects

    def insertIntoPapers(self,paperdata):
        """ Inserts into collection"""
        listoftitles = [paper.reference.title for paper in Paper.objects()]
        if paperdata['reference']['title'] in listoftitles:
            return None
        paper = Paper(**paperdata)
        paper.save()
        return str(paper.id)



    def __filtersearchedPaper(self, filteredPaper):
        """
        Produces search results after filter
        :return: list: filtered search objects
        """
        filteredSearchobjects = []
        for paper in filteredPaper:
            search = Search()
            search.id = str(paper.id)
            search.title = paper.reference.title
            search.tags = paper.tags
            search.collections = paper.collections
            search.authors = [authors.firstName + " " + authors.lastName for authors in paper.reference.authors]
            search.authors = ", ".join(search.authors)
            search.publication = paper.reference.journal.fullName + " " + paper.reference.volume + ", " + paper.reference.page
            search.abstract = paper.reference.publishedAbstract
            search.doi = paper.reference.DOI
            search.serverPath = paper.info.serverPath
            search.folderAbsolutePath = paper.info.folderAbsolutePath
            search.fileServerPath = paper.info.fileServerPath
            search.downloadPath = paper.info.downloadPath
            search.notebookPath = paper.info.notebookPath
            search.notebookFile = paper.info.notebookFile
            search.year = int(paper.reference.year)
            filteredSearchobjects.append(search.__dict__)
        return filteredSearchobjects

    def getPaperDetails(self, paperid):
        """
        Produces details of paper based on paper clicked
        :param paperid:
        :return: object: Details of paper
        """
        paperDetailsObject = Paper.objects.filter(id=str(paperid))
        paperDetails = PaperDetails()
        paper = paperDetailsObject[0]
        paperDetails.id = str(paper.id)
        paperDetails.title = paper.reference.title
        paperDetails.tags = paper.tags
        paperDetails.collections = paper.collections
        paperDetails.authors = [authors.firstName + " " + authors.lastName for authors in paper.reference.authors]
        paperDetails.authors = ", ".join(paperDetails.authors)
        paperDetails.publication = paper.reference.journal.fullName + " " + paper.reference.volume + ", " + paper.reference.page
        paperDetails.abstract = paper.reference.publishedAbstract
        paperDetails.doi = paper.reference.DOI
        paperDetails.serverPath = paper.info.serverPath
        paperDetails.folderAbsolutePath = paper.info.folderAbsolutePath
        paperDetails.fileServerPath = paper.info.fileServerPath
        paperDetails.downloadPath = paper.info.downloadPath
        paperDetails.notebookPath = paper.info.notebookPath
        paperDetails.notebookFile = paper.info.notebookFile
        paperDetails.year = int(paper.reference.year)
        paperDetails.charts = paper.charts
        paperDetails.datasets = paper.datasets
        paperDetails.scripts = paper.scripts
        paperDetails.tools = paper.tools
        paperDetails.workflows = paper.workflow
        paperDetails.heads = paper.heads
        paperDetails.cite = paper.info.doi
        paperDetails.timeStamp = paper.info.timeStamp
        return paperDetails.__dict__

    def getWorkflowDetails(self, paperid):
        """
        Build workflow details based on paper
        :param paperid: Id of paper to construct workflow
        :return: object: workflow details
        """
        try:
            paperDetailsObject = Paper.objects.filter(id=str(paperid))
            paper = paperDetailsObject[0]
            self.workflowinfo.paperTitle = paper.reference.title
            self.workflowinfo.edges = paper.workflow.edges
            for node in paper.workflow.nodes:
                self.__insertWorkflowNodeDetails(node, paper)
            self.workflowinfo.workflowType = "paper: " + paper.reference.title
            return self.workflowinfo.__dict__
        except Exception as e:
            print(e)


    def getWorkflowForChartDetails(self, paperid, chartid):
        """
        Build workflow details based on paper
        :param paperid: Id of paper to construct workflow
        :return: object: workflow details
        """
        paperDetailsObject = Paper.objects.filter(id=str(paperid))
        paper = paperDetailsObject[0]
        self.workflowinfo.paperTitle = paper.reference.title
        self.__insertWorkflowNodeDetails(chartid, paper)
        self.__addEdgeToWorkflowForChart(chartid, paper)
        return self.workflowinfo.__dict__

    def __addEdgeToWorkflowForChart(self, chartid, paper):
        """
        Adds nodes to build workflow for chart
        :return:
        """
        for edge in paper.workflow.edges:
            if chartid in edge[1]:
                self.workflowinfo.edges.append(edge)
                self.__addNodeToWorkflowForChart(paper, edge)
            else:
                self.hasvisited.append(edge[1])
                if self.__hasPath(edge[1], chartid, paper.workflow):
                    self.workflowinfo.edges.append(edge)
                    self.__addNodeToWorkflowForChart(paper, edge)
                self.hasvisited.clear()

    def __addNodeToWorkflowForChart(self, paper, edge):
        """
        Adds other node details for the chart
        :param paper:
        :param edge:
        :return:
        """
        if edge[0] not in self.workflowinfo.nodes:
            self.__insertWorkflowNodeDetails(edge[0], paper)
        elif edge[1] not in self.workflowinfo.nodes:
            self.__insertWorkflowNodeDetails(edge[1], paper)

    def __hasPath(self, destination, chartid, workflow):
        """
        Checks if path is found
        :param destination:
        :param chartid:
        :param workflow:
        :return:
        """
        for edge in workflow.edges:
            if destination in edge[0]:
                if edge[1] not in self.hasvisited:
                    self.hasvisited.append(edge[1])
                    if chartid in edge[1] or self.__hasPath(edge[1], chartid, workflow):
                        return True
        return False

    def __insertWorkflowNodeDetails(self, node, paper):
        """

        :param node:
        :param paper:
        :param nodeInfo:
        :return:
        """
        try:
            if "h" in node:
                for head in paper.heads:
                    if node in str(head.id):
                        workflownodeinfo = WorkflowNodeInfo()
                        workflownodeinfo.toolTip = "<p><b>External " + head.id + "</b>: <i>" + head.readme + "</i></p>"
                        details = []
                        details.append("Head " + node)
                        details.append("<p><i>" + head.readme + "</i></p>" + self.__getLinks(head.URLs))
                        workflownodeinfo.details = details
                        workflownodeinfo.hasNotebookFile = False
                        if head.saveas:
                            workflownodeinfo.nodelabel = head.saveas
                        else:
                            workflownodeinfo.nodelabel = node
                        self.workflowinfo.nodes[node] = workflownodeinfo.__dict__
            elif "t" in node:
                for tool in paper.tools:
                    if node in str(tool.id):
                        workflownodeinfo = WorkflowNodeInfo()
                        workflownodeinfo.toolTip = self.__getTooltipForTools(tool)
                        details = []
                        details.append("Tool " + node)
                        toollinks = self.__getLinks(tool.URLs)
                        tooldetails = self.__getTooltipForTools(tool)
                        toolfiles = self.__getFiles(tool.files, paper.info.fileServerPath)
                        workflowtools = tooldetails + toollinks + toolfiles + "</i></p>"
                        details.append(workflowtools)
                        workflownodeinfo.details = details
                        if tool.saveas:
                            workflownodeinfo.nodelabel = tool.saveas
                        else:
                            workflownodeinfo.nodelabel = node
                        workflownodeinfo.hasNotebookFile = False
                        self.workflowinfo.nodes[node] = workflownodeinfo.__dict__

            elif "d" in node:
                for dataset in paper.datasets:
                    if node in str(dataset.id):
                        workflownodeinfo = WorkflowNodeInfo()
                        workflownodeinfo.toolTip = self.__getTooltipForNode(dataset, node)
                        details = []
                        extradatasetfields = ""
                        details.append("Dataset " + node)
                        datasetlinks = self.__getLinks(dataset.URLs)
                        datasetdetails = self.__getTooltipForNode(dataset, node)
                        datasetfiles = self.__getFiles(dataset.files, paper.info.fileServerPath)
                        # extradatasetfields = self.__getExtraFields(dataset)
                        workflowdatasets = datasetdetails + datasetlinks + datasetfiles + extradatasetfields
                        details.append(workflowdatasets)
                        workflownodeinfo.details = details
                        if dataset.saveas:
                            workflownodeinfo.nodelabel = dataset.saveas
                        else:
                            workflownodeinfo.nodelabel = node
                        workflownodeinfo.hasNotebookFile = False
                        self.workflowinfo.nodes[node] = workflownodeinfo.__dict__

            elif "s" in node:
                for script in paper.scripts:
                    if node in str(script.id):
                        workflownodeinfo = WorkflowNodeInfo()
                        workflownodeinfo.toolTip = self.__getTooltipForNode(script, node)
                        details = []
                        extrascriptfields = ""
                        details.append("Script " + node)
                        scriptlinks = self.__getLinks(script.URLs)
                        scriptdetails = self.__getTooltipForNode(script, node)
                        scriptfiles = self.__getFiles(script.files, paper.info.fileServerPath)
                        workflowscripts = scriptdetails + scriptlinks + scriptfiles + extrascriptfields
                        details.append(workflowscripts)
                        workflownodeinfo.details = details
                        if script.saveas:
                            workflownodeinfo.nodelabel = script.saveas
                        else:
                            workflownodeinfo.nodelabel = node
                        workflownodeinfo.hasNotebookFile = False
                        self.workflowinfo.nodes[node] = workflownodeinfo.__dict__

            elif "c" in node:
                for chart in paper.charts:
                    if node in str(chart.id):
                        workflownodeinfo = WorkflowNodeInfo()
                        workflownodeinfo.toolTip = self.__getTooltipForNode(chart, node, paper.info.fileServerPath)
                        details = []
                        extrachartfields = ""
                        details.append("Chart " + node)
                        chartlinks = self.__getLinks(chart.properties, "charts")
                        chartdetails = self.__getTooltipForNode(chart, node, paper.info.fileServerPath)
                        chartfiles = self.__getFiles(chart.files, paper.info.fileServerPath)
                        # extrachartfields = self.__getExtraFields(chart)
                        workflowcharts = chartdetails + chartlinks + chartfiles + extrachartfields
                        details.append(workflowcharts)
                        workflownodeinfo.details = details
                        if chart.saveas:
                            workflownodeinfo.nodelabel = chart.saveas
                        else:
                            workflownodeinfo.nodelabel = node
                        workflownodeinfo.hasNotebookFile = False
                        self.workflowinfo.nodes[node] = workflownodeinfo.__dict__
                        self.workflowinfo.workflowType = chart.number + " of " + paper.reference.title
        except Exception as e:
            print(e)

    def __getLinks(self, urls, properties=None):
        links = ""
        if len(urls) > 0:
            if properties:
                for url in urls:
                    links = links + ", " + url
                if links:
                    links = "<p><b>Properties: </b>" + links.strip(",") + "</p>"
            else:
                for url in urls:
                    links = links + "<p><a href=" + url + " title=" + url + " target='_blank'>" + url + "</a></p>"
                if links:
                    links = links.strip()
                    links = links.strip(",")
                    links = "<p><b>URLs:</b>" + links + "</p>"
        return links

    def __getFiles(self, files, fileserverpath):
        filelinks = ""
        if len(files) > 0:
            for file in files:
                filename = file.rsplit('/', 1)[-1]
                if filename:
                    filelinks = filelinks + "<a href=" + fileserverpath + "/" + file + " title='Click to view " + filename + "' target='_blank'>" + filename + "</a>" + ", "
        if filelinks:
            filelinks = filelinks.strip()
            filelinks = filelinks.strip(",")
            filelinks = "<p><b>Files: </b>" + filelinks + "</p>"
        return filelinks

    def __getTooltipForTools(self, tool):
        details = ""
        if "experiment" in tool.kind:
            details = "<i>Experiment</i></p><p><b>Facility Name:</b> " + tool.facilityName + \
                      "</p><p><b>Measurement:</b> " + tool.measurement + "</p>"
        elif "software" in tool.kind:
            details = "<i>Software</i></p><p><b>Package Name:</b> " + tool.packageName \
                      + "</p><p><b>Executable Name:</b> " + tool.programName \
                      + "</p><p><b>Version:</b> " + tool.version
            if tool.readme:
                details = details + "</p><p><b>Readme:</b> " + tool.readme + "</p>"
        return "<p><b>Tool " + tool.id + "</b>: " + details

    def __getTooltipForNode(self, node, workflowtype, fileServerPath=None):
        tooltip = ""
        if "d" in workflowtype:
            tooltip = "<p><b>Dataset " + node.id + "</b>: <i>" + node.readme + "</i>"
        elif "s" in workflowtype:
            tooltip = "<p><b>Script " + node.id + "</b>: <i>" + node.readme + "</i>"
        elif "c" in workflowtype:
            tooltip = "<p><img src='" + fileServerPath + "/" + node.imageFile + "' class='img-responsive img-thumbnail'></p><p><b>" \
                      + node.number + ": </b><i>" + node.caption + "</i></p>"
        return tooltip

    def __getExtraFields(self, node):
        extraFieldValues = ""
        for hashkey, value in node.extraFields.items():
            if hashkey:
                for extrafieldkey, extrafieldval in hashkey.items():
                    extraFieldValues = extraFieldValues + "<p><b>" + extrafieldkey + ": </b><br>" + ", " + extrafieldval + "</p>"
        return extraFieldValues




from setuptools import setup,find_packages

setup(
    name='qresp',
    version='1.0.1',
    url='http://qresp.org/',
    entry_points = {
        'console_scripts': ['qresp=project.__main__:main'],
    },
    license='GNU',
    author='Aditya Tanikanti, Marco Govoni',
    author_email='datadev@lists.uchicago.edu',
    description='Qresp "Curation and Exploration of Reproducible Scientific Papers" is a Python application that facilitates the organization, annotation and exploration of data presented in scientific papers. ',
    python_requires='>=3.0',
    packages=find_packages(),
    install_requires=[
        'flask_api',
        'flask',
        'flask_cors',
        'paramiko',
        'pymongo',
        'cffi',
        'flask-mongoengine',
        'Flask-Session',
        'Flask-WTF',
        'mongoengine',
        'cryptography',
        'jinja2',
        'requests',
        'pyOpenSSL',
        'werkzeug',
        'itsdangerous',
        'python-dateutil',
        'expiringdict',
        'schedule',
        'wtforms',
        'flask-sitemap',
        'requests_oauthlib'
      ],
    include_package_data=True
)



import json
import re
import socket  # This method requires that we create our own socket
from copy import deepcopy
import paramiko  # Provides SSH functionality
import requests
from jsonschema import Draft4Validator
from requests_oauthlib import OAuth2Session


from .views import ReferenceForm, ChartForm, DatasetForm, ToolForm, ScriptForm, HeadForm





class DirectoryTree:
    """Class Providing Constants for Directory Tree.
    """
    title = ""
    parent = ""
    key = ""
    id = ""
    lazy = ""
    folder = ""
    source = ""


class Servers():
    """Class providing information about servers for federated search.
    """
    def __init__(self):
        self.__headers = {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.90 Safari/537.36'}
        self.__urlString = "https://paperstack.uchicago.edu/static/qresp_servers.json"
        self.__schemaString = "https://paperstack.uchicago.edu/static/v1_0.json"

    def getServersList(self):
        url = requests.get(self.__urlString, headers=self.__headers, verify = False)
        data = json.loads(url.text)
        return data

    def validateSchema(self,coll_data):
        exceptions = []
        try:
            url = requests.get(self.__schemaString, headers=self.__headers, verify = False)
            schema_coll_data = json.loads(url.text)
            a = Draft4Validator(schema_coll_data)
            for error in sorted(a.iter_errors(coll_data), key=str):
                print("errors, ",error)
                message = ""
                try:
                    exp = str(error.absolute_path[2]) + " is missing in " + str(error.absolute_path[0]) + " " + str(
                        error.absolute_path[1])
                    exceptions.append(exp)
                except IndexError:
                    try:
                        message = error.message
                        if "is not of type" or "is too short" in message:
                            message = "Missing values"
                        exp = str(message) + " in " + str(error.absolute_path[0]) + " "+ str(error.absolute_path[1])
                        exceptions.append(exp)
                    except IndexError:
                        try:
                            message = error.message
                            if "is not of type" or "is too short" in message:
                                message = "Missing values"
                            exp = str(message) + " in " + str(error.absolute_path[0])
                            exceptions.append(exp)
                        except IndexError:
                            exceptions.append(str(error.message))
        except:
            exceptions.append("Could not fetch schema")
        print(exceptions)
        return exceptions

class ConnectToServer():
    """  Class connecting to remote server with DUO login
    """
    def __init__(self,servername,username,password,isDUO,choice):
        # Global variables are used to store these data because they're sent to the server by a callback
        self.__user = username
        self.__pw = password
        self.__mfa = choice
        self.__serverName = servername
        self.__isDUO = isDUO
        self.__ftp = None
        self.connectToServer()

    def inter_handler(self,title, instructions, prompt_list):
        """
        Handler for remote server in curator
        :param title:
        :param instructions:
        :param prompt_list:
        :return:
        """

        resp = []  #Initialize the response container

        #Walk the list of prompts that the server sent that we need to answer
        for pr in prompt_list:
            #str() used to to make sure that we're dealing with a string rather than a unicode string
            #strip() used to get rid of any padding spaces sent by the server
            if str(pr[0]).strip() == "Username:":
                resp.append(self.__user)
            elif str(pr[0]).strip() == "Password:":
                resp.append(self.__pw)
            else:
                resp.append(self.__mfa)

        return tuple(resp)  #Convert the response list to a tuple and return it


    #Setup Paramiko logging; this is useful for troubleshooting

    #Get the username, password, and MFA token code from the user

    def call2FAsftpclient(self):

        #Create a socket and connect it to port 22 on the host
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.__serverName, 22))

        #Make a Paramiko Transport object using the socket
        ts = paramiko.Transport(sock)

        #Tell Paramiko that the Transport is going to be used as a client
        ts.start_client(timeout=10)

        #Begin authentication; note that the username and callback are passed
        ts.auth_interactive(self.__user, self.inter_handler)

        #Opening a session creates a channel along the socket to the server
        chan = ts.open_session(timeout=100)

        sftp = paramiko.SFTPClient.from_transport(ts)

        return sftp

    def connectToServer(self):
        if "No" in self.__isDUO:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(self.__serverName, username=self.__user, password=self.__pw)
            self.__ftp = ssh.open_sftp()
        else:
            self.__ftp = self.call2FAsftpclient()

    def getSftp(self):
        return self.__ftp

class Dtree():
    """ Class to build server tree
    """
    def __init__(self,ftp,path,session):
        self.__services = []
        self.__listObjects = []
        self.__ftp = ftp
        self.__path = path
        self.session = session

    def fetchForTree(self):
        for f in self.__ftp.listdir_attr(self.__path):
            dataFile = DirectoryTree()
            lstatout = str(f).split()[0]
            file = str(f).split()[8]
            if "qresp.ini" in file:
                self.openFileToReadConfig("qresp.ini")
            dataFile.title = file
            parent = self.__path.split("/")
            parentName = parent[len(parent) - 1]
            dataFile.key = self.__path + "/" + file
            dataFile.id = file
            relPath = str(self.__path + "/" + file).split(parentName, 1)[1]
            dataFile.parent = parentName + "/" + relPath
            if 'd' in lstatout:
                prev = file
                dataFile.folder = "true"
                dataFile.lazy = 'true'
            self.__listObjects.append(dataFile.__dict__)
        return self.__listObjects

    def openFileToReadConfig(self,configFile):
        try:
            remote_file = self.__ftp.open(str(self.__path + "/" + configFile))
            for line in remote_file:
                if str(line) and "=" in str(line):
                    linesplit = line.split("=", 1)
                    servicename = str(linesplit[0]).strip()
                    servicepath = str(linesplit[1]).strip()
                    if "http_service_path" in servicename:
                        if servicepath:
                            self.session["fileServerPath"] = servicepath
                            self.session["notebookPath"] = servicepath
                            self.__services.append("http_service_path")
                    elif "globus_service_path" in servicename:
                        if servicepath:
                            self.session["downloadPath"] = servicepath
                            self.__services.append("globus_service_path")
                    elif "isgitservice" in servicename:
                        if servicepath:
                            self.session["gitService"] = servicepath
                            self.__services.append("git_service")
        except:
            print("Config file not found")

    def checkIsConfigFile(self):
        ssh = ""
        isConfigFile = "N"
        try:
            for f in self.__ftp.listdir_attr(self.__path):
                dataFile = DirectoryTree()
                lstatout = str(f).split()[0]
                file = str(f).split()[8]
                if "qresp.ini" in file:
                    isConfigFile = "Y"
                    break
        except:
            print("Could not connect to remote server")
        return isConfigFile

    def fetchServices(self):
        return self.__services


class FetchDOI():
    """ Fetches information using DOI
    """
    def __init__(self,doi):
        self.__doi = doi

    def find_word(self,text, search):
        """Util Function to find exact match in words.
        :return: Information of the References - Dictionary
        """
        result = re.findall('\\b' + text + '\\b', search, flags=re.IGNORECASE)
        if len(result) > 0:
            return True
        else:
            return False

    def fetchFromDOI(self):
        """Collects and Returns values given DOI for the Reference section of the drop down menu of the Curation step.
        :return: Information of the References - Dictionary
        """
        try:
            headers = {
                'Accept': 'application/rdf+xml;q=0.5, application/vnd.citationstyles.csl+json;q=1.0',
            }
            res = requests.get('https://doi.org/' + str(self.__doi), headers=headers)
            res.encoding = 'utf-8'
            response = json.loads(res.text)
            kind = "Journal"
            title = ""
            journalFull = ""
            volume = ""
            page = ""
            year = 0
            publishedAbstract = ""
            url = ""
            personList = []
            for resp in response.keys():
                if self.find_word(str(resp).strip(), "title"):
                    title = str(response[resp])
                elif self.find_word(str(resp).strip(), "container-title"):
                    journalFull = str(response[resp])
                elif self.find_word(str(resp).strip(), "container-title-short"):
                    journalAbbr = str(response[resp])
                elif self.find_word(str(resp).strip(), "volume"):
                    volume = str(response[resp])
                elif self.find_word(str(resp).strip(), "article-number") or self.find_word(str(resp).strip().lower(), "page"):
                    page = str(response[resp])
                elif self.find_word(str(resp).strip(), "created"):
                    if "date-parts" in str(response['created']).strip():
                        if len(response['created']['date-parts']) > 0:
                            year = int(response['created']['date-parts'][0][0])
                    if "date-time" in str(response['created']).strip().lower():
                        publishedDate = str(response['created']['date-time'])
                elif self.find_word(str(resp).strip(), "abstract"):
                    publishedAbstract = str(response[resp])
                elif self.find_word(str(resp).strip(), "URL"):
                    url = str(response[resp])
                elif self.find_word(str(resp).strip(), "author"):
                    for eachAuth in response['author']:
                        person = {}
                        if "given" in eachAuth:
                            first = ""
                            middle = ""
                            if ' ' in str(eachAuth["given"]):
                                parts = str(eachAuth["given"]).split()
                                first = str(parts[0])
                                middle = str(parts[1])
                                person["firstName"] = first
                                person["middleName"] = middle
                            else:
                                first = str(eachAuth["given"])
                                person["firstName"] = first
                        if "family" in eachAuth:
                            person["lastName"] = str(eachAuth["family"])
                        personList.append(person)
            authors = {}
            authors["authors"] = personList
            ref = ReferenceForm(**authors)
            ref.journal.fullName.data = journalFull
            ref.URLs.data = url
            ref.publishedAbstract.data = publishedAbstract
            ref.year.data = year
            ref.page.data = page
            ref.volume.data = volume
            ref.title.data = title
            ref.DOI.data = self.__doi
        except Exception as e:
            print(e)
            raise IOError
        return ref


class GenerateIDs():
    """ Generates Ids to charts, datasets,scripts,heads
    """
    def __init__(self,session):
        self.session = session
        self.types = ["charts","tools","datasets","scripts","heads"]
        for type in self.types:
            listItems = session.get(type)
            if listItems:
                self.generateIDs(listItems,type)

    def generateIDs(self,listItems,type):
        formList = []
        form = None
        index = 0
        for item in listItems:
            if "charts" in type:
                form = ChartForm(**item)
            elif "tools" in type:
                form = ToolForm(**item)
            elif "datasets" in type:
                form = DatasetForm(**item)
            elif "scripts" in type:
                form = ScriptForm(**item)
            elif "heads" in type:
                form = HeadForm(**item)
            if not form.id.data:
                form.id.data = type[0] + str(index)
                index = index + 1
            formList.append(form.data)
        self.session[type] = deepcopy(formList)


class ConvertToList():
    """ Util method to convert URLs and files from string object to list object
    """
    def __init__(self,paper):
        self.converToList = []
        self.paper = paper

    def fetchConvertedList(self):
        paperdata = self.paper
        for chart in paperdata['charts']:
            if isinstance(chart['files'],str):
                chart['files'] = chart['files'].split(",")
            if isinstance(chart['properties'],str):
                chart['properties'] = chart['properties'].split(",")
        for tool in paperdata['tools']:
            if isinstance(tool['URLs'], str):
                tool['URLs'] = tool['URLs'].split(",")
            if isinstance(tool['patches'], str):
                tool['patches'] = tool['patches'].split(",")
        for dataset in paperdata['datasets']:
            if isinstance(dataset['files'], str):
                dataset['files'] = dataset['files'].split(",")
            if isinstance(dataset['URLs'], str):
                dataset['URLs'] = dataset['URLs'].split(",")
        for script in paperdata['scripts']:
            if isinstance(script['files'], str):
                script['files'] = script['files'].split(",")
            if isinstance(script['URLs'], str):
                script['URLs'] = script['URLs'].split(",")
        for head in paperdata['heads']:
            if isinstance(head['URLs'], str):
                head['URLs'] = head['URLs'].split(",")
        return paperdata


class SendDescriptor():
    """sends descriptor
    """
    def __init__(self,servername):
        self.__servername = servername

    def sendDescriptorToServer(self,data):
        """ Sends email to server
        :return: object: sends descriptor to server
        """
        url = self.__servername + "/getDescriptor"
        payload = {'metadata': data, 'servername': self.__servername}
        headers = {'Application': 'qresp', 'Accept': 'application/json', 'Content-Type': 'application/json'}
        response = requests.post(url, data=json.dumps(payload), headers=headers, verify=False)
        return response

class GoogleAuth():
    """ Authorization for google.
    """

    def __init__(self,clientid = None,redirecturi = None,scope = None):
        self.clientid = clientid
        self.redirecturi = redirecturi
        self.scope = scope


    def getGoogleAuth(self,state=None,token=None):
        if token:
            return OAuth2Session(self.clientid, token=token)
        if state:
            return OAuth2Session(
                self.clientid,
                state=state,
                redirect_uri=self.redirecturi)
        oauth = OAuth2Session(self.clientid,redirect_uri=self.redirecturi,scope=self.scope)
        return oauth